package example.graphicalEditor;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Point;
import java.awt.geom.Point2D;

import fr.lri.swingstates.canvas.CElement;
import fr.lri.swingstates.canvas.CExtensionalTag;
import fr.lri.swingstates.canvas.CRectangularShape;
import fr.lri.swingstates.canvas.CShape;
import fr.lri.swingstates.canvas.CStateMachine;
import fr.lri.swingstates.canvas.CTag;
import fr.lri.swingstates.canvas.Canvas;
import fr.lri.swingstates.sm.State;
import fr.lri.swingstates.sm.Transition;
import fr.lri.swingstates.sm.transitions.Click;
import fr.lri.swingstates.sm.transitions.Drag;
import fr.lri.swingstates.sm.transitions.Event;
import fr.lri.swingstates.sm.transitions.MouseOnPosition;
import fr.lri.swingstates.sm.transitions.Press;
import fr.lri.swingstates.sm.transitions.Release;
import fr.lri.swingstates.sm.transitions.TimeOut;

class SelectionTag extends CExtensionalTag {
	public void added(CShape s) {
		s.addGhost();
		s.fixReferenceShapeToCurrent();
		s.getGhost().setPickable(false).translateTo(s.getCenterX() - 3, s.getCenterY() - 3)
		.setOutlinePaint(Color.LIGHT_GRAY).setFilled(false).setTransparencyOutline(0.7f).setParent(s).below(s);
	}
	public void removed(CShape s) {
		s.removeGhost();
	}
};

/**
 * A standard Swing GUI graphical editor. This editor
 * is not fully functional, its interface has just been 
 * designed for illustration purpose.
 * 
 * @author Caroline Appert
 * @see example.player#StrokePlayer
 */
class GraphicalEditor extends Canvas {

	public static int TOOL_RECT = 1;
	public static int TOOL_ELL  = 2;
	private static int tool = TOOL_RECT;
//	private double lastMouseX = -1;
//	private double lastMouseY = -1;
	private double pasteX = -1;
	private double pasteY = -1;
	
	private CElement clipBoard;

	private DeSelectOrMove deSelectOrMove;
	private SelectOrMove selectOrMove;
	private DrawOrClick drawOrClick;
	
	

	public void setTool(int tool) {
		this.tool = tool;
	}

	class DragOrClick extends CStateMachine {
		protected double lastX, lastY;
		protected double firstX, firstY;
		protected int nbClicks;
		protected Canvas canvas;

		public DragOrClick(Canvas canvas) {
			super();
			nbClicks = 1;
			attachTo(canvas);
			this.canvas = canvas;
		}
		public DragOrClick(Canvas canvas, int nbClicks) {
			super();
			this.nbClicks = nbClicks;
			attachTo(canvas);
			this.canvas = canvas;
		}

		public State start = new State() { };

		public State pressed = new State() {
			Transition drag = new Drag(">> dragging") {
				public void action() {
					lastX = getPoint().getX();
					lastY = getPoint().getY();
				}
			};
			Transition release = new Release(">> maybeClick") { };
			Transition other = new MouseOnPosition(">> start") {
				public void action() {
					fireEvent("end");
				}
			};
		};

		public State maybeClick = new State() {
			public void enter() {
				armTimer(200, false);
			}
			Transition click = new Click(">> start") {
				public void action() {
					if(getMouseEvent().getClickCount() == nbClicks) click(this);
					fireEvent("end");
				}
			};
			Transition noClick = new TimeOut(">> start") {
				public void action() {
					fireEvent("end");
				}
			};
		};
		
		public State dragging = new State() {
			Transition drag = new Drag() {
				public void action() {
					drag(this);
					lastX = getPoint().getX();
					lastY = getPoint().getY();
				}
			};
			Transition other = new MouseOnPosition(">> start") {
				public void action() {
					fireEvent("end");
				}
			};
		};

		void drag(Drag transition) { }

		void click(Click transition) { }
	}

	class SelectOrMove extends DragOrClick {

		CElement moved = null;

		public SelectOrMove(Canvas canvas) {
			super(canvas, 1);
			start.addTransition(new PressOnShape(BUTTON1, NOMODIFIER, ">> pressed") {
				public void action() {
					moved = getShape();
					firstX = getPoint().getX();
					firstY = getPoint().getY();
					lastX = getPoint().getX();
					lastY = getPoint().getY();
					setMachineActive(SelectOrMove.this);
					consumes(true);
				}
			});
		}

		void drag(Drag transition) {
			moved.translateBy(transition.getPoint().getX() - lastX, transition.getPoint().getY() - lastY);
		}

		void click(Click transition) {
			moved.addTag(new SelectionTag());
		}

	}

	class DeSelectOrMove extends DragOrClick {

		CElement moved = null;
		SelectionTag selectionTag = null;

		public DeSelectOrMove(Canvas canvas) {
			super(canvas, 1);
			initStatesAndTransitions();
			start.addTransition(new PressOnTag(SelectionTag.class, BUTTON1, NOMODIFIER, ">> pressed") {
				public void action() {
					selectionTag = (SelectionTag)getTag();
					moved = GraphicalEditor.this.getTag(SelectionTag.class);
					firstX = getPoint().getX();
					firstY = getPoint().getY();
					lastX = getPoint().getX();
					lastY = getPoint().getY();
					setMachineActive(DeSelectOrMove.this);
					consumes(true);
				}
			});
		}

		void drag(Drag transition) {
			moved.translateBy(transition.getPoint().getX() - lastX, transition.getPoint().getY() - lastY);
		}

		void click(Click transition) {
			moved.removeTag(selectionTag);
		}

	}

	void setMachineActive(CStateMachine machine) {
		deSelectOrMove.setActive(deSelectOrMove == machine);
		selectOrMove.setActive(selectOrMove == machine);
		drawOrClick.setActive(drawOrClick == machine);
	}

	void setAllMachinesActive() {
		deSelectOrMove.setActive(true);
		selectOrMove.setActive(true);
		drawOrClick.setActive(true);
	}

	class DrawOrClick extends DragOrClick {

		CRectangularShape drawn;
		double pathLength = 0;
		
		public DrawOrClick(Canvas canvas) {
			super(canvas, 2);
			initStatesAndTransitions();
			start.addTransition(new Press(BUTTON1, ">> pressed") {
				public void action() {
					drawn= null;
					pathLength = 0;
					firstX = getPoint().getX();
					firstY = getPoint().getY();
					lastX = getPoint().getX();
					lastY = getPoint().getY();
					setMachineActive(DrawOrClick.this);
					consumes(true);
				}
			});
		}

		void drag(Drag transition) {
			if(drawn == null) {
				if(tool == TOOL_ELL) {
					drawn = newEllipse(firstX, firstY, 1, 1);
				} else {
					// tool == TOOL_RECT
					drawn = newRectangle(firstX, firstY, 1, 1);
				}
				Color randomColor = new Color((float)Math.random(), (float)Math.random(), (float)Math.random());
				drawn.setOutlinePaint(randomColor).setFillPaint(randomColor).setStroke(new BasicStroke(3)).setTransparencyFill(0.3f);
			} else {
				drawn.setDiagonal(firstX, firstY, lastX, lastY);
			}
		}

	}

	CStateMachine interaction = new CStateMachine() {

		public void init() {
			deSelectOrMove = new DeSelectOrMove(GraphicalEditor.this);
			deSelectOrMove.addStateMachineListener(this);
			selectOrMove = new SelectOrMove(GraphicalEditor.this);
			selectOrMove.addStateMachineListener(this);
			drawOrClick = new DrawOrClick(GraphicalEditor.this);
			drawOrClick.addStateMachineListener(this);
		}

		State start = new State() {
//			Transition mouse = new MouseOnPosition() {
//				public void action() {
//					lastMouseX = getPoint().getX();
//					lastMouseY = getPoint().getY();
//				}
//			};
			Transition end = new Event("end") {
				public void action() {
					setAllMachinesActive();
				}
			};
		};
	};
	
	public GraphicalEditor(int width, int height) {
		super(width, height);
		interaction.attachTo(this);
	}

	public void cut() {
		clipBoard = getTag(SelectionTag.class).setTransparencyOutline(0.3f);
		clipBoard.addTag("toCut");
		copy();
	}

	public void copy() {
		CTag tg = getTag(SelectionTag.class);
		CExtensionalTag copied = new CExtensionalTag() { };
		tg.reset();
		while(tg.hasNext()) {
			CShape copy = tg.nextShape().duplicate();
			copy.addTag(copied);
		}
		copied.setDrawable(false).setPickable(false);
		if(clipBoard != null) clipBoard.setTransparencyOutline(1f);
		clipBoard = copied;
		pasteX = clipBoard.getCenterX() + 10;
		pasteY = clipBoard.getCenterY() + 10;
	}

	public void paste() {
//		if(clipBoard == null && lastMouseX != -1) return;
//		clipBoard.translateBy(lastMouseX - clipBoard.getCenterX(), lastMouseY - clipBoard.getCenterY());
		if(clipBoard == null) return;
		clipBoard.translateBy(pasteX - clipBoard.getCenterX(), pasteY - clipBoard.getCenterY());
		
		clipBoard.setTransparencyOutline(1f).setDrawable(true).setPickable(true);
		CTag toCut = getTag("toCut");
		if(toCut != null) removeShapes(toCut);
		copy();
	}

	public void clear() {
		removeAllShapes();		
	}
	
	public void setPastePosition(double pasteX, double pasteY) {
		this.pasteX = pasteX;
		this.pasteY = pasteY;
	}
	
	public void setPastePosition(Point pt) {
		this.pasteX = pt.getX();
		this.pasteY = pt.getY();
	}
	
	public void setPastePosition(Point2D pt) {
		this.pasteX = pt.getX();
		this.pasteY = pt.getY();
	}
}
